import fs from 'node:fs';
import stream from 'node:stream';
import type { Readable } from 'node:stream';
import type AWS from '@aws-sdk/client-s3';
import { S3, type S3ClientConfig } from '@aws-sdk/client-s3';
import { DataStore, Upload, type KvStore } from '@tus/utils';
import { Semaphore } from '@shopify/semaphore';
export type Options = {
    /**
     * The preferred part size for parts send to S3. Can not be lower than 5MiB or more than 5GiB.
     * The server calculates the optimal part size, which takes this size into account,
     * but may increase it to not exceed the S3 10K parts limit.
     */
    partSize?: number;
    /**
     * The minimal part size for parts.
     * Can be used to ensure that all non-trailing parts are exactly the same size.
     * Can not be lower than 5MiB or more than 5GiB.
     */
    minPartSize?: number;
    /**
     * The maximum number of parts allowed in a multipart upload. Defaults to 10,000.
     */
    maxMultipartParts?: number;
    useTags?: boolean;
    maxConcurrentPartUploads?: number;
    cache?: KvStore<MetadataValue>;
    expirationPeriodInMilliseconds?: number;
    s3ClientConfig: S3ClientConfig & {
        bucket: string;
    };
};
export type MetadataValue = {
    file: Upload;
    'upload-id': string;
    'tus-version': string;
};
export declare class S3Store extends DataStore {
    protected bucket: string;
    protected cache: KvStore<MetadataValue>;
    protected client: S3;
    protected preferredPartSize: number;
    protected expirationPeriodInMilliseconds: number;
    protected useTags: boolean;
    protected partUploadSemaphore: Semaphore;
    maxMultipartParts: number;
    minPartSize: number;
    maxUploadSize: 5497558138880;
    constructor(options: Options);
    protected shouldUseExpirationTags(): boolean;
    protected useCompleteTag(value: 'true' | 'false'): string | undefined;
    /**
     * Saves upload metadata to a `${file_id}.info` file on S3.
     * Please note that the file is empty and the metadata is saved
     * on the S3 object's `Metadata` field, so that only a `headObject`
     * is necessary to retrieve the data.
     */
    protected saveMetadata(upload: Upload, uploadId: string): Promise<void>;
    protected completeMetadata(upload: Upload): Promise<void>;
    /**
     * Retrieves upload metadata previously saved in `${file_id}.info`.
     * There's a small and simple caching mechanism to avoid multiple
     * HTTP calls to S3.
     */
    protected getMetadata(id: string): Promise<MetadataValue>;
    protected infoKey(id: string): string;
    protected partKey(id: string, isIncomplete?: boolean): string;
    protected uploadPart(metadata: MetadataValue, readStream: fs.ReadStream | Readable, partNumber: number): Promise<string>;
    protected uploadIncompletePart(id: string, readStream: fs.ReadStream | Readable): Promise<string>;
    protected downloadIncompletePart(id: string): Promise<{
        size: number;
        path: string;
        createReader: (options: {
            cleanUpOnEnd: boolean;
        }) => fs.ReadStream;
    } | undefined>;
    protected getIncompletePart(id: string): Promise<Readable | undefined>;
    protected getIncompletePartSize(id: string): Promise<number | undefined>;
    protected deleteIncompletePart(id: string): Promise<void>;
    /**
     * Uploads a stream to s3 using multiple parts
     */
    protected uploadParts(metadata: MetadataValue, readStream: stream.Readable, currentPartNumber: number, offset: number): Promise<number>;
    /**
     * Completes a multipart upload on S3.
     * This is where S3 concatenates all the uploaded parts.
     */
    protected finishMultipartUpload(metadata: MetadataValue, parts: Array<AWS.Part>): Promise<string | undefined>;
    /**
     * Gets the number of complete parts/chunks already uploaded to S3.
     * Retrieves only consecutive parts.
     */
    protected retrieveParts(id: string, partNumberMarker?: string): Promise<Array<AWS.Part>>;
    /**
     * Removes cached data for a given file.
     */
    protected clearCache(id: string): Promise<void>;
    protected calcOptimalPartSize(size?: number): number;
    /**
     * Creates a multipart upload on S3 attaching any metadata to it.
     * Also, a `${file_id}.info` file is created which holds some information
     * about the upload itself like: `upload-id`, `upload-length`, etc.
     */
    create(upload: Upload): Promise<Upload>;
    read(id: string): Promise<stream.Readable>;
    /**
     * Write to the file, starting at the provided offset
     */
    write(src: stream.Readable, id: string, offset: number): Promise<number>;
    getUpload(id: string): Promise<Upload>;
    declareUploadLength(file_id: string, upload_length: number): Promise<void>;
    remove(id: string): Promise<void>;
    protected getExpirationDate(created_at: string): Date;
    getExpiration(): number;
    deleteExpired(): Promise<number>;
    protected uniqueTmpFileName(template: string): Promise<string>;
}
//# sourceMappingURL=index.d.ts.map